
#include "CalibrationAlg.h"

#include <sstream>
#include <fstream>
#include <string>
#include <cmath>

using std::cout;
using std::endl;
using std::vector;
using std::string;
using std::pow;

CalibrationAlg::CalibrationAlg(Event* event, Geometry* geo, Headers* runHeader):SystemAlg( event,geo,runHeader){
  m_event = event;
  m_geo = geo;
  m_runHeader = runHeader;
  // hardwired number of calibration coefficients (this info is not in the calibration file)
  m_hecNumberOfCoeff = 4;
  m_emecNumberOfCoeff = 4;
}


// initialize any variables here
bool CalibrationAlg::initialize(){

  return true;
}

// put any code you want executed per event here
bool CalibrationAlg::execute(){

  return true;
}

//put any code here that you want executed after all the events have been processed.
bool CalibrationAlg::finalize(){

  return true;
}

//===== hec public methods =====

// Load hec calibration coefficients from file
bool CalibrationAlg::hecLoad(string calibFileName){

  // open the file
  ifstream fStream(calibFileName.c_str());

  // exit if problems opening the file
  if(!fStream.is_open()){
    cout << "Failed to open file:" << calibFileName << endl;
    exit(1);
  }

  // get the calibration file version number
  fStream >> m_hecVersion;

  while (!fStream.eof()) {
    int febno = 0;
    vector<float> coeff;   // m_hecNumberOfCoeff elements set to 0
    for (int i = 0; i < m_hecNumberOfCoeff; ++i) coeff.push_back(0);
    int status = 0;
    //    fStream >> febno >> coeff[0] >> coeff[1] >> coeff[2] >> coeff[3] >> status;
    fStream >> febno;
    for (int i = 0; i < m_hecNumberOfCoeff; ++i) {
      fStream >> coeff[i];
    }
    fStream >> status;
    if (!fStream.eof()) {
      if (m_geo->hec_isConnected(febno)) {
        // fill the map of febno keyed hec calibration coefficient and status
        m_hecCoeffMap[febno] = coeff;
        m_hecCoeffStatusMap[febno] = status;
        // fill the map of febno keyed hec pedestal rms in nA
        m_hecPedRMSnAMap[febno] = (m_runHeader->hec()->getPedRms(febno)) * coeff[1];
      }
    }
  }

  fStream.close();

  // check for corrupted calibration file
  // the maps should contain all the connected channels
  // it turns out the hec calibration files have entries for all valid febno, connected or not
  if (m_hecCoeffMap.size() != m_geo->hec_LocalCenterMap().size()) {
    cout << "**** wrong number of connected channels in calibration file " << calibFileName 
         << "  : " << m_hecCoeffMap.size() << endl;
    exit(1);
  }

  cout << "Read calibration file " << calibFileName << " version " << m_hecVersion << " for " 
       << m_hecCoeffMap.size() << " connected channels" << endl;

  return true;
}

// Get febno keyed hec calibration coefficient map
map<int, vector<float> > CalibrationAlg::hecCoeffMap(){
  return m_hecCoeffMap;
}

// Get febno keyed hec calibration coefficient status map
map<int, int> CalibrationAlg::hecCoeffStatusMap(){
  return m_hecCoeffStatusMap;
}

// Get hec calibration coefficients
vector<float> CalibrationAlg::hecCoeff(int febno){
  if (m_geo->hec_isConnected(febno)) {
    return m_hecCoeffMap[febno];
  } else {
    cout << "febno not connected in hecCoeff(): " << febno << endl;
    exit(1);
  }
}

// Get hec calibration coefficient status
int CalibrationAlg::hecCoeffStatus(int febno){
  if (m_geo->hec_isConnected(febno)) {
    return m_hecCoeffStatusMap[febno];
  } else {
    cout << "febno not connected in hecCoeffStatus(): " << febno << endl;
    exit(1);
  }
}

// Get febno keyed hec ped rms in nA map
map<int, float> CalibrationAlg::hecPedRMSnAMap(){
  return m_hecPedRMSnAMap;
}

// Get hec ped rms in nA
float CalibrationAlg::hecPedRMSnA(int febno){
  if (m_geo->hec_isConnected(febno)) {
    return m_hecPedRMSnAMap[febno];
  } else {
    cout << "febno not connected in hecPedRMSnA(): " << febno << endl;
    exit(1);
  }
}

// pedestal subtracted ADC to nA conversion for the hec
float CalibrationAlg::hecADCTonA(int febno, float ADC){
  if (m_geo->hec_isConnected(febno)) {
    float nA = 0;
    for (int i = 1; i < m_hecNumberOfCoeff; ++i) {
      nA += hecCoeff(febno)[i] * pow(ADC, i);
    }
    return nA;
  } else {
    cout << "febno not connected in hecADCTonA(): " << febno << endl;
    exit(1);
  }
}


//===== emec public methods =====

// Load emec calibration coefficients from file
bool CalibrationAlg::emecLoad(string calibFileName){

  // open the file
  ifstream fStream(calibFileName.c_str());

  // exit if problems opening the file
  if(!fStream.is_open()){
    cout << "Failed to open file:" << calibFileName << endl;
    exit(1);
  }

  // get the calibration file version number
  fStream >> m_emecVersion;

  while (!fStream.eof()) {
    int febno = 0;
    vector<float> coeff;   // m_emecNumberOfCoeff elements set to 0
    for (int i = 0; i < m_emecNumberOfCoeff; ++i) coeff.push_back(0);
    int status = 0;
    //    fStream >> febno >> coeff[0] >> coeff[1] >> coeff[2] >> coeff[3] >> status;
    fStream >> febno;
    for (int i = 0; i < m_emecNumberOfCoeff; ++i) {
      fStream >> coeff[i];
    }
    fStream >> status;
    if (!fStream.eof()) {
      if (m_geo->emec_isConnected(febno)) {
        // fill the map of febno keyed emec calibration coefficient and status
        m_emecCoeffMap[febno] = coeff;
        m_emecCoeffStatusMap[febno] = status;
        // fill the map of febno keyed emec pedestal rms in nA
        m_emecPedRMSnAMap[febno] = (m_runHeader->emec()->getPedRms(febno)) * coeff[1];
      }
    }
  }

  fStream.close();

  // check for corrupted calibration file
  // the maps should contain all the connected channels
  // it turns out the emec calibration files have entries up to the last connected channel
  if (m_emecCoeffMap.size() != m_geo->emec_LocalCenterMap().size()) {
    cout << "**** wrong number of connected channels in calibration file " << calibFileName 
         << "  : " << m_emecCoeffMap.size() << endl;
    exit(1);
  }

  cout << "Read calibration file " << calibFileName << " version " << m_emecVersion << " for " 
       << m_emecCoeffMap.size() << " connected channels" << endl;
 
  return true;
}

// Get febno keyed emec calibration coefficient map
map<int, vector<float> > CalibrationAlg::emecCoeffMap(){
  return m_emecCoeffMap;
}

// Get febno keyed emec calibration coefficient status map
map<int, int> CalibrationAlg::emecCoeffStatusMap(){
  return m_emecCoeffStatusMap;
}

// Get emec calibration coefficients
vector<float> CalibrationAlg::emecCoeff(int febno){
  if (m_geo->emec_isConnected(febno)) {
    return m_emecCoeffMap[febno];
  } else {
    cout << "febno not connected in emecCoeff(): " << febno << endl;
    exit(1);
  }
}

// Get emec calibration coefficient status
int CalibrationAlg::emecCoeffStatus(int febno){
  if (m_geo->emec_isConnected(febno)) {
    return m_emecCoeffStatusMap[febno];
  } else {
    cout << "febno not connected in emecCoeffStatus(): " << febno << endl;
    exit(1);
  }
}

// Get febno keyed emec ped rms in nA map
map<int, float> CalibrationAlg::emecPedRMSnAMap(){
  return m_emecPedRMSnAMap;
}

// Get emec ped rms in nA
float CalibrationAlg::emecPedRMSnA(int febno){
  if (m_geo->emec_isConnected(febno)) {
    return m_emecPedRMSnAMap[febno];
  } else {
    cout << "febno not connected in emecPedRMSnA(): " << febno << endl;
    exit(1);
  }
}

// pedestal subtracted ADC to nA conversion for the emec
float CalibrationAlg::emecADCTonA(int febno, float ADC){
  if (m_geo->emec_isConnected(febno)) {
    float nA = 0;
    for (int i = 1; i < m_emecNumberOfCoeff; ++i) {
      nA += emecCoeff(febno)[i] * pow(ADC, i);
    }
    return nA;
  } else {
    cout << "febno not connected in emecADCTonA(): " << febno << endl;
    exit(1);
  }
}


