
#include "OccupancyAlg.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <cmath>
#include <vector>

#include "TF1.h"

using std::cout;
using std::setw;
using std::endl;
using std::vector;

OccupancyAlg::OccupancyAlg(Event* event, Geometry* geo, Headers* runHeader):SystemAlg( event,geo,runHeader){
  m_event = event;
  m_geo = geo;
  m_runHeader = runHeader;

  // create the new System Alg Objects
  m_noise = new NoiseAlg(m_event, m_geo, m_runHeader);
  m_calibration = new CalibrationAlg(m_event, m_geo, m_runHeader);
  vector<SystemAlg*> algVector;

  algVector.push_back(m_noise);
  algVector.push_back(m_calibration);
  SystemAlg::setAlgs(algVector);

  // occupancy computation enable flag
  m_hecEnable = false;
  m_emecEnable = false;
  // signal cut variables
  m_hecNoiseCut = 2.;  //  |signal| > m_hecNoiseCut*Noise
  m_emecNoiseCut = 2.;  //  |signal| > m_emecNoiseCut*Noise
  // print level
  m_hecPrintLevel = 1;
  m_emecPrintLevel = 1;
  // occupancy result file name
  m_hecOccupancyFileName = "";
  m_emecOccupancyFileName = "";
  // calibration file name
  m_hecCalibrationFileName = "";
  m_emecCalibrationFileName = "";
  // event counters
  m_hecEvents = 0;
  m_emecEvents = 0;
}


// initialize any variables here
bool OccupancyAlg::initialize(){
  SystemAlg::initialize();

  return true;
}

// put any code you want executed per event here
bool OccupancyAlg::execute(){
  SystemAlg::execute();

  if (!m_hecEnable && !m_emecEnable) return true;

  // only keep physics trigger events
  // Triggers (physics,electron,muon,pi,random)
  // Particle (fELECTRON=1, fMUON=2, fPION=3)
  if (m_event->getHecTriggerBit(0) == 0 || m_event->getHecTriggerBit(4) == 1) {  // not physics or a random
    return true;
  }

  // in the case of global cubic fit timing, only keep good events
  if (m_runHeader->hec()->getTimingType() > 0) { // global cubic time
    if (m_event->globalCubicTimeSigma() == 0) { // bad global timing
      return true;
    }
  }

  // fill occupancy maps
  // if data in ADC, convert to nA
  if (m_hecEnable) {
    for (int febno = m_runHeader->hec()->getFirstCh(); febno <= m_runHeader->hec()->getLastCh(); ++febno) {
      if (m_geo->hec_isConnected(febno)) {
        float s = m_event->hecCell(febno)->signal();
        if (m_runHeader->hec()->getEUnit() == 1) s = m_calibration->hecADCTonA(febno, s);  // ADC to nA
        if (fabs(s) > m_hecNoiseCut*m_noise->hecNoise(febno)) ++m_hecOccupancyMap[febno];
      }
    }
    ++m_hecEvents;
  }
  if (m_emecEnable) {
    for (int febno = m_runHeader->emec()->getFirstCh(); febno <= m_runHeader->emec()->getLastCh(); ++febno) {
      if (m_geo->emec_isConnected(febno)) {
        float s = m_event->emecCell(febno)->signal();
        if (m_runHeader->emec()->getEUnit() == 1) s = m_calibration->emecADCTonA(febno, s);  // ADC to nA
        if (fabs(s) > m_emecNoiseCut*m_noise->emecNoise(febno)) ++m_emecOccupancyMap[febno];
      }
    }
    ++m_emecEvents;
  }

  return true;
}

//put any code here that you want executed after all the events have been processed.
bool OccupancyAlg::finalize(){
  SystemAlg::finalize();//do NOT remove this line

  if (m_hecEnable) {
    // fill the occupancy maps
    // should not iterate over the map in case a channel has no entries (0 occupancy)
    for (int febno = m_runHeader->hec()->getFirstCh(); febno <= m_runHeader->hec()->getLastCh(); ++febno) {
      if (m_geo->hec_isConnected(febno)) {
        if (m_hecEvents > 0) {
          m_hecOccupancyMap[febno] /= m_hecEvents;
          // binomial error
          float var = m_hecOccupancyMap[febno] * (1. - m_hecOccupancyMap[febno]) / m_hecEvents;
          m_hecOccupancyErrorMap[febno] = (var > 0.) ? sqrt(var) : 0.;
        } else {
          m_hecOccupancyMap[febno] = 0.;  // create the map entry
          m_hecOccupancyErrorMap[febno] = 0.;  // create the map entry
        }
      }
    }
    if (m_hecPrintLevel >= 1) hecPrintOccupancy();
    // write occupancy data to file
    hecWriteOccupancyFile();
  }

  if (m_emecEnable) {
    // fill the occupancy maps
    // should not iterate over the map in case a channel has no entries (0 occupancy)
    for (int febno = m_runHeader->emec()->getFirstCh(); febno <= m_runHeader->emec()->getLastCh(); ++febno) {
      if (m_geo->emec_isConnected(febno)) {
        if (m_emecEvents > 0) {
          m_emecOccupancyMap[febno] /= m_emecEvents;
          // binomial error
          float var = m_emecOccupancyMap[febno] * (1. - m_emecOccupancyMap[febno]) / m_emecEvents;
          m_emecOccupancyErrorMap[febno] = (var > 0.) ? sqrt(var) : 0.;
        } else {
          m_emecOccupancyMap[febno] = 0.;  // create the map entry
          m_emecOccupancyErrorMap[febno] = 0.;  // create the map entry
        }
      }
    }
    if (m_emecPrintLevel >= 1) emecPrintOccupancy();
    // write occupancy data to file
    emecWriteOccupancyFile();
  }

  return true;
}

//################################
// hec public methods

// enable the computation of channel occupancy for the hec
void OccupancyAlg::hecEnable(string noiseFileName){
  // set the enable flag
  m_hecEnable = true;
  if (noiseFileName == "") {
    // get the default file name
    // hec_nnnnn_noise.dat
    std::ostringstream os;
    os << "hec_" << m_runHeader->global()->getRunNo() << "_noise.dat";
    noiseFileName = os.str();
  }
  // load the noise file
  m_noise->hecReadNoiseFile(noiseFileName);
  // calibration file must be given if data is in ADC
  if (m_runHeader->hec()->getEUnit() == 1) { // ADC
    if (m_hecCalibrationFileName == "") {
      cout << "The ntuple data is in ADC: you must specify a calibration file with hecCalibrationFileName" << endl;
      exit(1);
    } else {  // load the calibration coefficients
      m_calibration->hecLoad(m_hecCalibrationFileName);
    }
  }
}

// Get the hec occupancy
float OccupancyAlg::hecOccupancy(int febno){
  if (m_geo->hec_isConnected(febno)) {
    return m_hecOccupancyMap[febno];
  } else {
    cout << "febno not connected in hecOccupancy(): " << febno << endl;
    exit(1);
  }
}

// Get the hec occupancy error estimate
float OccupancyAlg::hecOccupancyError(int febno){
  if (m_geo->hec_isConnected(febno)) {
    return m_hecOccupancyErrorMap[febno];
  } else {
    cout << "febno not connected in hecOccupancyError(): " << febno << endl;
    exit(1);
  }
}

// Read a hec occupancy file
void OccupancyAlg::hecReadOccupancyFile(string occupancyFileName){
  // see hecWriteOccupancyFile() for the file format
  if (occupancyFileName == "") {
    // get the default file name
    // hec_nnnnn_occupancy.dat
    std::ostringstream os;
    os << "hec_" << m_runHeader->global()->getRunNo() << "_occupancy.dat";
    occupancyFileName = os.str();
  }
  // open the file
  ifstream fStream(occupancyFileName.c_str());
  // exit if problems opening the file
  if(!fStream.is_open()){
    cout << "hecReadOccupancyFile failed to open file:" << occupancyFileName << endl;
    exit(1);
  }
  //
  int runNumber, idet;
  float m_hecNoiseCut;
  fStream >> runNumber;
  fStream >> idet;
  fStream >> m_hecNoiseCut;
  if (idet != 2) { // idet = 2 for the hec
    cout << "Occupancy file does not contain hec data " << occupancyFileName << endl;
    exit(1);
  }
  // clear the occupancy maps in case they have already been produced or read
  m_hecOccupancyMap.clear();
  m_hecOccupancyErrorMap.clear();
  // read the rest of the occupancy file
  while (!fStream.eof()) {
    int febno = 0;
    float occupancy = 0.;
    float occupancyError = 0.;
    fStream >> febno >> occupancy >> occupancyError;
    if (!fStream.eof()) {
      if (m_geo->hec_isConnected(febno)) {
        // fill the map of febno keyed hec occupancy
        m_hecOccupancyMap[febno] = occupancy;
        m_hecOccupancyErrorMap[febno] = occupancyError;
      }
    }
  }
  fStream.close();
  cout << "hec occupancy values from run " << runNumber 
       << " made with noise cut " << m_hecNoiseCut
       << " sigmas read from file " << occupancyFileName 
       << " for " << m_hecOccupancyMap.size() << " channels" << endl;
}

// Print hec occupancy data
void OccupancyAlg::hecPrintOccupancy(){
  cout << "hec occupancy" << endl;
  cout << "   febno     occupancy" << endl;
  for (map<int, float>::iterator it = m_hecOccupancyMap.begin(); it != m_hecOccupancyMap.end(); ++it) {
    int febno = it->first;
    cout << setw(8)  << febno
         << setw(12) << m_hecOccupancyMap[febno] << " +/-"
         << setw(12) << m_hecOccupancyErrorMap[febno] << endl;
  }
}

//################################
// emec public methods

// enable the computation of channel occupancy for the emec
void OccupancyAlg::emecEnable(string noiseFileName){
  // set the enable flag
  m_emecEnable = true;
  if (noiseFileName == "") {
    // get the default file name
    // emech_nnnnn_noise.dat or emecm_nnnnn_noise.dat or emec_nnnnn_noise.dat
    // since the gain is not recorded in the ntuple (!) use the data file name
    string s;
    s.assign(m_runHeader->emec()->getDataFileName(), 0, 5);  // get first 5 characters
    if (s != "emech" && s != "emecm") s = "emec";
    std::ostringstream os;
    os << s+"_" << m_runHeader->global()->getRunNo() << "_noise.dat";
    noiseFileName = os.str();
  }
  // load the noise file
  m_noise->emecReadNoiseFile(noiseFileName);
  // calibration file must be given if data is in ADC
  if (m_runHeader->emec()->getEUnit() == 1) { // ADC
    if (m_emecCalibrationFileName == "") {
      cout << "The ntuple data is in ADC: you must specify a calibration file with emecCalibrationFileName" << endl;
      exit(1);
    } else {  // load the calibration coefficients
      m_calibration->emecLoad(m_emecCalibrationFileName);
    }
  }
}

// Get the emec occupancy
float OccupancyAlg::emecOccupancy(int febno){
  if (m_geo->emec_isConnected(febno)) {
    return m_emecOccupancyMap[febno];
  } else {
    cout << "febno not connected in emecOccupancy(): " << febno << endl;
    exit(1);
  }
}

// Get the emec occupancy error estimate
float OccupancyAlg::emecOccupancyError(int febno){
  if (m_geo->emec_isConnected(febno)) {
    return m_emecOccupancyErrorMap[febno];
  } else {
    cout << "febno not connected in emecOccupancyError(): " << febno << endl;
    exit(1);
  }
}

// Read a emec occupancy file
void OccupancyAlg::emecReadOccupancyFile(string occupancyFileName){
  // see emecWriteOccupancyFile() for the file format
  if (occupancyFileName == "") {
    // get the default file name
    // emech_nnnnn_occupancy.dat or emecm_nnnnn_occupancy.dat or emec_nnnnn_occupancy.dat
    // since the gain is not recorded in the ntuple (!) use the data file name
    string s;
    s.assign(m_runHeader->emec()->getDataFileName(), 0, 5);  // get first 5 characters
    if (s != "emech" && s != "emecm") s = "emec";
    std::ostringstream os;
    os << s+"_" << m_runHeader->global()->getRunNo() << "_occupancy.dat";
    occupancyFileName = os.str();
  }
  // open the file
  ifstream fStream(occupancyFileName.c_str());
  // exit if problems opening the file
  if(!fStream.is_open()){
    cout << "emecReadOccupancyFile failed to open file:" << occupancyFileName << endl;
    exit(1);
  }
  //
  int runNumber, idet;
  float m_emecNoiseCut;
  fStream >> runNumber;
  fStream >> idet;
  fStream >> m_emecNoiseCut;
  if (idet != 1) { // idet = 1 for the emec
    cout << "Occupancy file does not contain emec data " << occupancyFileName << endl;
    exit(1);
  }
  // clear the occupancy maps in case they have already been produced or read
  m_emecOccupancyMap.clear();
  m_emecOccupancyErrorMap.clear();
  // read the rest of the occupancy file
  while (!fStream.eof()) {
    int febno = 0;
    float occupancy = 0.;
    float occupancyError = 0.;
    fStream >> febno >> occupancy >> occupancyError;
    if (!fStream.eof()) {
      if (m_geo->emec_isConnected(febno)) {
        // fill the map of febno keyed emec occupancy
        m_emecOccupancyMap[febno] = occupancy;
        m_emecOccupancyErrorMap[febno] = occupancyError;
      }
    }
  }
  fStream.close();
  cout << "emec occupancy values from run " << runNumber 
       << " made with noise cut " << m_emecNoiseCut
       << " sigmas read from file " << occupancyFileName 
       << " for " << m_emecOccupancyMap.size() << " channels" << endl;
}

// Print emec occupancy data
void OccupancyAlg::emecPrintOccupancy(){
  cout << "emec occupancy" << endl;
  cout << "   febno     occupancy" << endl;
  for (map<int, float>::iterator it = m_emecOccupancyMap.begin(); it != m_emecOccupancyMap.end(); ++it) {
    int febno = it->first;
    cout << setw(8)  << febno
         << setw(12) << m_emecOccupancyMap[febno] << " +/-"
         << setw(12) << m_emecOccupancyErrorMap[febno] << endl;
  }
}

//################################
// hec private methods

// write occupancy data file
void OccupancyAlg::hecWriteOccupancyFile(){
  // format is
  // line 1:      run number
  // line 2:      idet
  // line 3:      m_hecNoiseCut
  // other lines: febno  occupancy  error
  //
  if (m_hecOccupancyFileName == "") {
    // get the default file name
    // hec_nnnnn_occupancy.dat
    std::ostringstream os;
    os << "hec_" << m_runHeader->global()->getRunNo() << "_occupancy.dat";
    m_hecOccupancyFileName = os.str();
  }
  // open the file
  ofstream fStream(m_hecOccupancyFileName.c_str());
  // exit if problems opening the file
  if(!fStream.is_open()){
    cout << "hecWriteOccupancyFile failed to open file:" << m_hecOccupancyFileName << endl;
    exit(1);
  }
  //
  fStream << m_runHeader->global()->getRunNo() << endl;
  fStream << 2 << endl;  // idet = 2 for the hec
  fStream << m_hecNoiseCut << endl;
  for (map<int, float>::iterator it = m_hecOccupancyMap.begin(); it != m_hecOccupancyMap.end(); ++it) {
    int febno = it->first;
    fStream << setw(8)  << febno
            << setw(15) << m_hecOccupancyMap[febno]
            << setw(15) << m_hecOccupancyErrorMap[febno] << endl;
  }
  fStream.close();
  cout << "hec occupancy values written in file " << m_hecOccupancyFileName << endl;
}

//################################
// emec private methods

// write occupancy data file
void OccupancyAlg::emecWriteOccupancyFile(){
  // format is
  // line 1:      run number
  // line 2:      idet
  // line 3:      m_emecNoiseCut
  // other lines: febno  occupancy  error
  //
  if (m_emecOccupancyFileName == "") {
    // get the default file name
    // emech_nnnnn_occupancy.dat or emecm_nnnnn_occupancy.dat or emec_nnnnn_occupancy.dat
    // since the gain is not recorded in the ntuple (!) use the data file name
    string s;
    s.assign(m_runHeader->emec()->getDataFileName(), 0, 5);  // get first 5 characters
    if (s != "emech" && s != "emecm") s = "emec";
    std::ostringstream os;
    os << s+"_" << m_runHeader->global()->getRunNo() << "_occupancy.dat";
    m_emecOccupancyFileName = os.str();
  }
  // open the file
  ofstream fStream(m_emecOccupancyFileName.c_str());
  // exit if problems opening the file
  if(!fStream.is_open()){
    cout << "emecWriteOccupancyFile failed to open file:" << m_emecOccupancyFileName << endl;
    exit(1);
  }
  //
  fStream << m_runHeader->global()->getRunNo() << endl;
  fStream << 1 << endl;  // idet = 1 for the emec
  fStream << m_emecNoiseCut << endl;
  for (map<int, float>::iterator it = m_emecOccupancyMap.begin(); it != m_emecOccupancyMap.end(); ++it) {
    int febno = it->first;
    fStream << setw(8)  << febno
            << setw(15) << m_emecOccupancyMap[febno]
            << setw(15) << m_emecOccupancyErrorMap[febno] << endl;
  }
  fStream.close();
  cout << "emec occupancy values written in file " << m_emecOccupancyFileName << endl;
}

   
